from fastapi import APIRouter, Request, File, UploadFile, Header
from typing import Optional
import numpy as np
from fastapi.responses import JSONResponse
from icecream import ic


router = APIRouter()


@router.post("/detect")
async def http_detect(request: Request, file: UploadFile = File(...), mac: str = Header(...), tags: Optional[str] = Header(None)):
    """
    Example Usage

        curl --request POST -H 'mac: FF:FF' -F 'file=@test.jpg' localhost:8004/detect

        for i in `seq 1 10`; do time curl --request POST -H 'mac: FF:FF' -F 'file=@test.jpg' localhost:8004/detect; done
    """
    try:
        np_image = np.fromfile(file.file, np.uint8)
    except:
        raise Exception(f'Uploaded file {file.filename} is no image file.')
    detections = await request.app.get_detections(np_image, mac, tags.split(',') if tags else [])
    return JSONResponse(detections)



from learning_loop_node.detector import BoxDetection, PointDetection, Detections
from learning_loop_node import Detector, ModelInformation


class TestingDetector(Detector):
    __test__ = False

    def __init__(self) -> None:
        super().__init__('mocked')

    def init(self,  model_info: ModelInformation, model_root_path: str):
        pass

    def evaluate(self, image: bytes) -> Detections:
        return Detections(
            box_detections=[BoxDetection('some_category_name', 1, 2, 3, 4, 'some_model', 42)],
            point_detections=[PointDetection('some_category_name', 10, 12, 'some_model', 42)]
        )




import asyncio
import functools
import logging
from learning_loop_node.socket_response import SocketResponse
from icecream import ic


def try_except(func):
    @functools.wraps(func)
    async def wrapper_handle_error(*args, **kwargs):
        try:
            if asyncio.iscoroutinefunction(func):
                value = await func(*args, **kwargs)
            else:
                value = func(*args, **kwargs)
            if isinstance(value, str):
                return SocketResponse.for_success(value).__dict__
                
            return value
        except Exception as e:
            logging.error(f'An error occured for {func.__name__}: {str(e)}')
            return SocketResponse.for_failure(str(e)).__dict__
            
    return wrapper_handle_error




from typing import Optional
from pydantic.main import BaseModel
from enum import Enum


class CategoryType(str, Enum):
    Box = 'box'
    Point = 'point'
    Segmentation = 'segmentation'


class Category(BaseModel):
    id: str
    name: str
    description: Optional[str]
    hotkey: Optional[str]
    color: Optional[str]
    type: Optional[CategoryType] = CategoryType.Box


def create_category(id: str, name: str, type: CategoryType):
    return Category(id=id, name=name, description='', hotkey='', color='', type=type)



import asyncio
import os
from typing import Dict, List, Optional, Union
from uuid import uuid4
from .executor import Executor
from .training import Training
from .model import BasicModel
from ..context import Context
from ..node import Node
from .downloader import TrainingsDownloader
from ..rest import downloads, uploads
from .. import node_helper
import logging
from icecream import ic


class Trainer():

    def __init__(self, model_format: str) -> None:
        self.model_format: str = model_format
        self.training: Optional[Training] = None
        self.executor: Optional[Executor] = None

    async def begin_training(self, context: Context, source_model: dict) -> None:
        downloader = TrainingsDownloader(context)

        self.training = Trainer.generate_training(context, source_model)
        self.training.data = await downloader.download_training_data(self.training.images_folder)
        self.executor = Executor(self.training.training_folder)
        logging.info(f'downloading model {source_model["id"]} as {self.model_format}')
        await downloads.download_model(self.training.training_folder, context, source_model['id'], self.model_format)
        logging.info(f'now starting training')
        await self.start_training()

    async def start_training(self) -> None:
        raise NotImplementedError()

    def stop_training(self) -> bool:
        if self.executor:
            self.executor.stop()
            return True
        else:
            logging.info('could not stop training, executor is None')
            return False

    def get_error(self) -> Optional[Union[None, str]]:
        '''Should be used to provide error informations to the Learning Loop by extracting data from self.executor.get_log().'''
        pass

    def get_log(self) -> str:
        return self.executor.get_log()

    async def save_model(self,  context: Context, model_id: str) -> None:
        files = await asyncio.get_running_loop().run_in_executor(None, self.get_model_files, model_id)
        if isinstance(files, list):
            await uploads.upload_model(context, files, model_id, self.model_format)
        elif isinstance(files, dict):
            for format in files:
                await uploads.upload_model(context, files[format], model_id, format)
        else:
            raise TypeError(f'can only save model as list or dict, but was {files}')

    def get_new_model(self) -> Optional[BasicModel]:
        '''Is called frequently to check if a new "best" model is availabe.
        Returns None if no new model could be found. Otherwise BasicModel(confusion_matrix, meta_information).
        `confusion_matrix` contains a dict of all classes: 
            - The classes must be identified by their id, not their name.
            - For each class a dict with tp, fp, fn is provided (true positives, false positives, false negatives).
        `meta_information` can hold any data which is helpful for self.on_model_published to store weight file etc for later upload via self.get_model_files
        '''
        raise NotImplementedError()

    def on_model_published(self, basic_model: BasicModel, model_id: str) -> None:
        '''Called after a BasicModel has been successfully send to the Learning Loop.
        The model_id is an uuid to identify the model within the Learning Loop.
        self.get_model_files uses this id to gather all files needed for transfering the actual data from the trainer node to the Learning Loop.
        In the simplest implementation this method just renames the weight file (encoded in BasicModel.meta_information) into a file name containing the model_id.
        '''
        raise NotImplementedError()

    def get_model_files(self, model_id) -> Union[List[str], Dict[str, List[str]]]:
        '''Called when the Learning Loop requests to backup a specific model. 
        Should return a list of file paths which describe the model.
        These files must contain all data neccessary for the trainer to resume a training (eg. weight file, hyperparameters, etc.) 
        and will be stored in the Learning Loop unter the format of this trainer.
        Note: by convention the weightfile should be named "model.<extension>" where extension is the file format of the weightfile.
        For example "model.pt" for pytorch or "model.weights" for darknet/yolo.

        If a trainer can also generate other formats (for example for an detector),
        a dictionary mapping format -> list of files can be returned.
        In general, all non-trainer formats are expected to contain a model.json in the file list which contains the trained resolution, categories etc.
        Example: {"resolution": 832, "categories":[{"name": "A", "id": "<a uuid>", "type": "box"}]}
        '''
        raise NotImplementedError()

    @staticmethod
    def generate_training(context: Context, source_model: dict) -> Training:
        training_uuid = str(uuid4())
        project_folder = Node.create_project_folder(context)
        return Training(
            id=training_uuid,
            base_model=source_model,
            context=context,
            project_folder=project_folder,
            images_folder=node_helper.create_image_folder(project_folder),
            training_folder=Trainer.create_training_folder(project_folder, training_uuid)
        )

    @staticmethod
    def create_training_folder(project_folder: str, trainings_id: str) -> str:
        training_folder = f'{project_folder}/trainings/{trainings_id}'
        os.makedirs(training_folder, exist_ok=True)
        return training_folder



from learning_loop_node.context import Context
import logging
from learning_loop_node.globals import GLOBALS
from .status import Status, State
from fastapi import FastAPI
import socketio
import asyncio
import asyncio
import os
from icecream import ic
from .loop import loop
from fastapi_utils.tasks import repeat_every
import logging
from uuid import uuid4
from learning_loop_node.decorators.error_handler import try_except
from datetime import datetime


class Node(FastAPI):
    name: str
    uuid: str

    def __init__(self, name: str, uuid: str = None):
        super().__init__()
        host = os.environ.get('LOOP_HOST', None) or os.environ.get('HOST', 'learning-loop.ai')
        self.ws_url = f'ws{"s" if host != "backend" else ""}://' + host

        self.name = name
        self.uuid = self.read_or_create_uuid() if uuid is None else uuid
        self.startup_time = datetime.now()

        self.sio_client = socketio.AsyncClient(
            reconnection_delay=0,
            request_timeout=0.5,
            # logger=True, engineio_logger=True
        )
        self.sio_client._trigger_event = try_except(self.sio_client._trigger_event)
        self.reset()

        @self.sio_client.on('connect')
        async def on_connect():
            logging.debug('received "on_connect" from constructor event.')
            self.reset()
            state = self.get_state()
            await self.update_state(state)

        @self.sio_client.on('disconnect')
        async def on_disconnect():
            logging.debug('received "on_disconnect" from constructor event.')
            await self.update_state(State.Offline)

        self.register_lifecycle_events()

    def read_or_create_uuid(self) -> str:
        if not os.path.exists(f'{GLOBALS.data_folder}/uuid.txt'):
            os.makedirs(GLOBALS.data_folder, exist_ok=True)
            uuid = str(uuid4())
            with open(f'{GLOBALS.data_folder}/uuid.txt', 'a+') as f:
                f.write(uuid)
        else:
            with open(f'{GLOBALS.data_folder}/uuid.txt', 'r') as f:
                uuid = f.read()

        return uuid

    def register_lifecycle_events(self):
        @self.on_event("startup")
        async def startup():
            logging.debug('received "startup" event')

        @self.on_event("shutdown")
        async def shutdown():
            logging.debug('received "shutdown" event')
            await self.sio_client.disconnect()

        @self.on_event("startup")
        @repeat_every(seconds=10, raise_exceptions=False, wait_first=False)
        async def ensure_connected() -> None:
            if not self.sio_client.connected:
                await self.connect()

    def reset(self):
        self.status = Status(id=self.uuid, name=self.name)

    async def get_state(self):
        raise Exception("Override this in subclass")

    async def connect(self):
        try:
            await self.sio_client.disconnect()
        except:
            pass

        logging.info(f'connecting to Learning Loop at {self.ws_url}')
        try:
            headers = await asyncio.get_event_loop().run_in_executor(None, loop.get_headers)
            await self.sio_client.connect(f"{self.ws_url}", headers=headers, socketio_path="/ws/socket.io")
            logging.debug(f'my sid is {self.sio_client.sid}')
            logging.info('connected to Learning Loop')
        except socketio.exceptions.ConnectionError as e:
            logging.error(f'socket.io connection error to "{self.ws_url}"')
        except Exception:
            logging.exception(f'error while connecting to "{self.ws_url}"')

    async def update_state(self, state: State):
        self.status.state = state
        if self.status.state != State.Offline:
            await self.send_status()

    async def update_status(self, new_status: Status):
        self.status.id = new_status.id
        self.status.name = new_status.name
        self.status.uptime = new_status.uptime
        self.status.latest_error = new_status.latest_error

        if self.status.state != State.Offline:
            self.status.state = State.Idle
        await self.send_status()

    async def send_status(self):
        raise Exception("Override this in subclass")

    @staticmethod
    def create_project_folder(context: Context) -> str:
        project_folder = f'{GLOBALS.data_folder}/{context.organization}/{context.project}'
        os.makedirs(project_folder, exist_ok=True)
        return project_folder



import os
from glob import glob
from fastapi.encoders import jsonable_encoder
import os
from learning_loop_node.detector import Detections
from typing import List
import json
from datetime import datetime
import requests
import logging
import shutil
from ..globals import GLOBALS


class Outbox():

    def __init__(self) -> None:
        self.path = f'{GLOBALS.data_folder}/outbox'
        if not os.path.exists(self.path):
            os.makedirs(self.path)

        host = os.environ.get('HOST', 'learning-loop.ai')
        base: str = f'http{"s" if host != "backend" else ""}://' + host
        o = os.environ.get('ORGANIZATION')
        p = os.environ.get('PROJECT')
        self.target_uri = f'{base}/api/{o}/projects/{p}/images'
        self.upload_in_progress = False

    def save(self, image: bytes, detections: Detections = Detections(), tags: List[str] = []) -> None:
        id = datetime.now().isoformat(sep='_', timespec='milliseconds')
        tmp = f'{GLOBALS.data_folder}/tmp/{id}'
        detections.tags = tags
        detections.date = id
        os.makedirs(tmp, exist_ok=True)
        with open(tmp + '/image.json', 'w') as f:
            json.dump(jsonable_encoder(detections), f)

        with open(tmp + '/image.jpg', 'wb') as f:
            f.write(image)
        os.rename(tmp, self.path + '/' + id)  # NOTE rename is atomic so upload can run in parallel
        from icecream import ic
        ic(self.path + '/' + id)

    def get_data_files(self):
        return glob(f'{self.path}/*')

    def upload(self) -> None:
        if self.upload_in_progress:
            return
        self.upload_in_progress = True

        try:
            for item in self.get_data_files():
                data = [('files', open(f'{item}/image.json', 'r')),
                        ('files', open(f'{item}/image.jpg', 'rb'))]

                response = requests.post(self.target_uri, files=data)
                if response.status_code == 200:
                    shutil.rmtree(item)
                    logging.info(f'uploaded {item} successfully')
                elif response.status_code == 422:
                    logging.error(f'Broken content in {item}: dropping this data')
                    shutil.rmtree(item)
                else:
                    logging.error(f'Could not upload {item}: {response.status_code}, {response.content}')
        except:
            logging.exception('could not upload files')
        self.upload_in_progress = False



from typing import List
from dataclasses import dataclass, field

from learning_loop_node.detector.box_detection import BoxDetection
from learning_loop_node.detector.point_detection import PointDetection


@dataclass
class Detections:
    box_detections: List[BoxDetection] = field(default_factory=list)
    point_detections: List[PointDetection] = field(default_factory=list)



from pydantic.main import BaseModel
from typing import List, Optional
from learning_loop_node.context import Context
from learning_loop_node.data_classes.category import Category


class ModelInformation(BaseModel):
    id: str
    host: str
    organization: str
    project: str
    version: str
    categories: List[Category]
    resolution: Optional[int]

    @property
    def context(self):
        return Context(organization=self.organization, project=self.project)



from .rest.operation_mode import OperationMode
from .detector import Detector
from .rest import detect, upload, operation_mode
from ..rest import downloads
from ..node import Node
from ..status import State
from ..status import DetectionStatus, State
from ..context import Context
from fastapi.encoders import jsonable_encoder
from fastapi_utils.tasks import repeat_every
from typing import List, Optional, Union
import shutil
import os
import contextlib
import logging
from learning_loop_node.globals import GLOBALS
import subprocess
import asyncio
import numpy as np
from fastapi_socketio import SocketManager
from . import Detections
from . import Outbox
from threading import Thread
from datetime import datetime
from icecream import ic


class DetectorNode(Node):
    update_frequency = 10

    def __init__(self, name: str, detector: Detector, uuid: str = None):
        super().__init__(name, uuid)
        self.detector = detector
        self.organization = os.environ.get('LOOP_ORGANIZATION', None) or os.environ.get('ORGANIZATION', None)
        self.project = os.environ.get('LOOP_PROJECT', None) or os.environ.get('PROJECT', None)
        assert self.organization, 'Detector node needs an organization'
        assert self.project, 'Detector node needs an project'
        self.operation_mode = OperationMode.Startup
        self.connected_clients = []
        self.outbox = Outbox()
        self.target_model = None
        self.include_router(detect.router, tags=["detect"])
        self.include_router(upload.router, prefix="")
        self.include_router(operation_mode.router, tags=["operation_mode"])

        @self.on_event("startup")
        @repeat_every(seconds=self.update_frequency, raise_exceptions=False, wait_first=False)
        async def _check_for_update() -> None:
            await self.check_for_update()

        @self.on_event("startup")
        async def _load_model() -> None:
            try:
                self.detector.load_model()
            finally:
                self.operation_mode = OperationMode.Idle

        @self.on_event("startup")
        @repeat_every(seconds=30, raise_exceptions=False, wait_first=False)
        def submit() -> None:
            thread = Thread(target=self.outbox.upload)
            thread.start()

        sio = SocketManager(app=self)

        @self.sio.on("detect")
        async def _detect(sid, data) -> None:
            try:
                np_image = np.frombuffer(data['image'], np.uint8)
                return await self.get_detections(np_image, data.get('mac', None), data.get('tags', []), data.get('active_learning', True))
            except Exception as e:
                logging.exception('could not detect via socketio')
                with open('/tmp/bad_img_from_socket_io.jpg', 'wb') as f:
                    f.write(data['image'])
                return {'error': str(e)}

        @self.sio.on("info")
        async def _info(sid) -> None:
            if self.detector.current_model:
                return self.detector.current_model.__dict__
            return 'No model loaded'

        @self.sio.on('upload')
        async def _upload(sid, data):
            loop = asyncio.get_event_loop()
            try:
                await loop.run_in_executor(None, lambda: self.outbox.save(data['image'], Detections(), ['picked_by_system']))
            except Exception as e:
                logging.exception('could not upload via socketio')
                return {'error': str(e)}

        @self.sio.event
        def connect(sid, environ, auth):
            self.connected_clients.append(sid)

        @self.on_event("shutdown")
        async def shutdown():
            for sid in self.connected_clients:
                await self.sio.disconnect(sid)

    async def check_for_update(self):
        if self.operation_mode == OperationMode.Startup:
            return
        try:
            ic()
            logging.info(f'periodically checking operation mode. Current mode is {self.operation_mode}')
            update_to_model_id = await self.send_status()
            if self.detector.current_model:
                logging.info(
                    f'Current model: {self.detector.current_model.version} with id {self.detector.current_model.id}')
            else:
                logging.info(f'no model loaded')
            if self.operation_mode != OperationMode.Idle:
                logging.info(f'not checking for updates; operation mode is {self.operation_mode}')
                return
            if self.target_model is None:
                logging.info(f'not checking for updates; no target model selected')
                return
            logging.info('going to check for new updates')
            if not self.detector.current_model or self.target_model != self.detector.current_model.version:
                logging.info(
                    f'Current model "{self.detector.current_model.version if self.detector.current_model else "-"}" needs to be updated to {self.target_model}')
                with pushd(GLOBALS.data_folder):
                    model_symlink = 'model'
                    target_model_folder = f'models/{self.target_model}'
                    shutil.rmtree(target_model_folder, ignore_errors=True)
                    os.makedirs(target_model_folder)
                    try:
                        await downloads.download_model(
                            target_model_folder,
                            Context(organization=self.organization, project=self.project),
                            update_to_model_id,
                            self.detector.model_format
                        )
                        try:
                            os.unlink(model_symlink)
                            os.remove(model_symlink)
                        except:
                            pass
                        os.symlink(target_model_folder, model_symlink)
                        logging.info(f'Updated symlink for model to {os.readlink(model_symlink)}')
                        self.reload(because='new model installed')
                    except downloads.DownloadError as e:
                        self.status.latest_error = f'download failed: {e.cause}'
            else:
                logging.info('Versions are identic. Nothing to do.')
        except Exception:
            logging.exception(f'check_for_update failed')
            self.status.latest_error = 'could not check for model update'

    async def send_status(self) -> Union[str, bool]:
        if not self.sio_client.connected:
            logging.error('could not send status -- we are not connected to the Learning Loop')
            return False
        status = DetectionStatus(
            id=self.uuid,
            name=self.name,
            state=self.status.state,
            latest_error=self.status.latest_error,
            uptime=int((datetime.now() - self.startup_time).total_seconds()),
            operation_mode=self.operation_mode,
            current_model=self.detector.current_model.version if self.detector.current_model else None,
            target_model=self.target_model,
            model_format=self.detector.model_format,
        )
        logging.debug(f'sending status {status}')
        response = await self.sio_client.call('update_detector', (self.organization, self.project, jsonable_encoder(status)), timeout=1)
        try:
            self.target_model = response['payload']['target_model_version']
            logging.debug(f'After sending status. Target_model is {self.target_model}')
            return response['payload']['target_model_id']
        except:
            logging.error('Could not send status to loop')
            return False

    def get_state(self):
        return State.Online

    async def set_operation_mode(self, mode: OperationMode):
        self.operation_mode = mode
        await self.send_status()

    def reload(self, because: str):
        '''provide a cause for the reload'''

        print('########## reloading app because ' + because, flush=True)
        if os.path.isfile('/app/restart/restart.py'):
            subprocess.call(['touch', '/app/restart/restart.py'])
        else:
            subprocess.call(['touch', '/app/main.py'])

    async def get_detections(self, raw_image, mac: str, tags: str, active_learning=True):
        loop = asyncio.get_event_loop()
        # image = await loop.run_in_executor(None, lambda: cv2.imdecode(np_image, cv2.IMREAD_COLOR))
        detections = await loop.run_in_executor(None, self.detector.evaluate, raw_image)
        info = "\n    ".join([str(d) for d in detections.box_detections])
        logging.info(f'detected:\n    {info}')
        # if active_learning:
        #     thread = Thread(target=learn, args=(detections, mac, tags, image))
        #     thread.start()
        return jsonable_encoder(detections)

    async def upload_images(self, images: List[bytes]):
        loop = asyncio.get_event_loop()
        for image in images:
            await loop.run_in_executor(None, lambda: self.outbox.save(image, Detections(), ['picked_by_system']))


@contextlib.contextmanager
def pushd(new_dir):
    previous_dir = os.getcwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(previous_dir)



import pytest
from learning_loop_node import DetectorNode
import requests
import json
import os
from glob import glob


def test_rest_detect(test_detector_node: DetectorNode):
    image = {('file', open('detector/tests/test.jpg', 'rb'))}
    headers = {'mac': '0:0:0:0', 'tags':  'some_tag'}
    request = requests.post(f'http://localhost:{pytest.detector_port}/detect', files=image, headers=headers)
    assert request.status_code == 200

    json_content = json.loads(request.content.decode('utf-8'))
    assert len(json_content['box_detections']) == 1
    assert json_content['box_detections'][0]['category_name'] == 'some_category_name'


@pytest.mark.asyncio
async def test_sio_detect(test_detector_node: DetectorNode, sio_client):
    with open('detector/tests/test.jpg', 'rb') as f:
        image_bytes = f.read()
    result = await sio_client.call('detect', {'image': image_bytes})
    assert len(result['box_detections']) == 1
    assert result['box_detections'][0]['category_name'] == 'some_category_name'


def test_rest_detect(test_detector_node: DetectorNode):
    image = {('file', open('detector/tests/test.jpg', 'rb'))}
    headers = {'mac': '0:0:0:0', 'tags':  'some_tag'}
    response = requests.post(f'http://localhost:{pytest.detector_port}/detect', files=image, headers=headers)
    assert response.status_code == 200
    result = response.json()
    assert len(result['box_detections']) == 1
    assert result['box_detections'][0]['category_name'] == 'some_category_name'


def test_rest_upload(test_detector_node: DetectorNode):
    assert test_detector_node.outbox.path.startswith('/tmp')

    def get_outbox_files():
        files = glob(f'{test_detector_node.outbox.path}/**/*', recursive=True)
        return [file for file in files if os.path.isfile(file)]

    assert len(get_outbox_files()) == 0
    image = {('files', open('detector/tests/test.jpg', 'rb'))}
    response = requests.post(f'http://localhost:{pytest.detector_port}/upload', files=image)
    assert response.status_code == 200
    assert len(get_outbox_files()) == 2, 'There should be one image and one .json file.'


@pytest.mark.asyncio
async def test_sio_upload(test_detector_node: DetectorNode, sio_client):
    assert test_detector_node.outbox.path.startswith('/tmp')

    def get_outbox_files():
        files = glob(f'{test_detector_node.outbox.path}/**/*', recursive=True)
        return [file for file in files if os.path.isfile(file)]

    assert len(get_outbox_files()) == 0

    with open('detector/tests/test.jpg', 'rb') as f:
        image_bytes = f.read()
    result = await sio_client.call('upload', {'image': image_bytes})
    assert result == None

    assert len(get_outbox_files()) == 2, 'There should be one image and one .json file.'



from typing import Optional, Union, Any
import os
import json
import logging
from learning_loop_node.model_information import ModelInformation
from learning_loop_node.globals import GLOBALS
from learning_loop_node.detector.detections import Detections


class Detector():

    def __init__(self, model_format: str) -> None:
        self.model_format: str = model_format
        self.current_model: Optional[ModelInformation] = None
        self.target_model: Optional[str] = None

    def load_model(self):
        try:
            model_root_path = f'{GLOBALS.data_folder}/model'
            model_info_file_path = f'{model_root_path}/model.json'
            if not os.path.exists(model_info_file_path):
                raise FileExistsError(f"File '{model_info_file_path}' does not exist.")
            with open(model_info_file_path, 'r') as f:
                try:
                    content = json.load(f)
                except:
                    raise Exception(f"could not read model information from file '{model_info_file_path}'")
                try:
                    model_information = ModelInformation.parse_obj(content)
                except Exception as e:
                    raise Exception(
                        f"could not parse model information from file '{model_info_file_path}'. \n {str(e)}")
            try:
                self.init(model_information, model_root_path)
                self.current_model = model_information
                logging.info(f'Successfully loaded model {self.current_model}')
            except Exception:
                logging.error(f'Could not init model {model_information}')
                raise

        except Exception:
            self.current_model = None
            logging.exception('An error occured during loading model.')

    def init(self,  model_info: ModelInformation, model_root_path: str):
        raise NotImplementedError()

    def evaluate(self, image: Any) -> Detections:
        raise NotImplementedError()



from typing import List, Optional
from pydantic import BaseModel
from enum import Enum
from learning_loop_node.context import Context


class State(str, Enum):
    Idle = "idle"
    Offline = "offline"
    Online = "online"
    Running = "running"
    Preparing = "preparing"


class Status(BaseModel):
    id: str
    name: str
    state: Optional[State] = State.Offline
    uptime: Optional[int] = 0
    latest_error: Optional[str] = None


class AnnotationNodeStatus(Status):
    capabilities: List[str]


class TrainingStatus(Status):
    latest_produced_model_id: Optional[str]
    train_image_count: Optional[int]
    test_image_count: Optional[int]
    skipped_image_count: Optional[int]


class DetectionStatus(Status):
    model_format: str
    current_model: Optional[str]
    target_model: Optional[str]
    operation_mode: Optional[str]



from typing import Optional
from learning_loop_node.context import Context
from learning_loop_node.rest.downloader import DataDownloader
from learning_loop_node.task_logger import create_task
from learning_loop_node.trainer.training_data import TrainingData
from learning_loop_node.data_classes.basic_data import BasicData
import logging
import os
from icecream import ic


class TrainingsDownloader():
    downloader: DataDownloader
    data_query_params: str

    def __init__(self, context: Context, data_query_params: Optional[str] = 'state=complete'):
        self.data_query_params = data_query_params
        self.downloader = DataDownloader(context)

    async def download_training_data(self, image_folder: str) -> TrainingData:
        basic_data = await self.downloader.download_basic_data(query_params=self.data_query_params)
        training_data = await self.download_images_and_annotations(basic_data, image_folder)
        return training_data

    async def download_images_and_annotations(self, basic_data: BasicData, image_folder: str) -> TrainingData:
        await self.downloader.download_images(basic_data.image_ids, image_folder)
        image_data = await self.downloader.download_images_data(basic_data.image_ids)
        logging.info('filtering corrupt images')  # download only safes valid images
        training_data = TrainingData(image_data=[], categories={c['name']: c['id'] for c in basic_data.categories})
        for i in image_data:
            if os.path.isfile(f'{image_folder}/{i["id"]}.jpg'):
                training_data.image_data.append(i)
            else:
                training_data.skipped_image_count += 1
        logging.info(f'Done downloading image data for {len(image_data)} images.')
        return training_data
